#include "../berLame_src/CallStack.h"
#include "../berLame_src/Dir.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

char *p_s_ReadFile(const char *p_s_filename)
{
	FILE *p_fr;
	if(!(p_fr = fopen(p_s_filename, "rb")))
		return 0;
	int n_length;
	char *p_s_file;
	if(fseek(p_fr, 0, SEEK_END) || (n_length = ftell(p_fr)) < 0 ||
	   fseek(p_fr, 0, SEEK_SET) || !(p_s_file = new char[n_length + 1])) {
		fclose(p_fr);
		return 0;
	}
	if(fread(p_s_file, sizeof(char), n_length, p_fr) != n_length) {
		delete[] p_s_file;
		fclose(p_fr);
		return 0;
	}
	fclose(p_fr);
	p_s_file[n_length] = 0;
	return p_s_file;
}

int n_Line(const char *p_s_text, const char *p_s_where)
{
	if(!p_s_where || p_s_where < p_s_text)
		return 0;
	int n_line = 1;
	while(p_s_text != p_s_where) {
		if(*p_s_text == '\n')
			++ n_line;
		++ p_s_text;
	}
	return n_line;
}

int n_Column(const char *p_s_text, const char *p_s_where)
{
	if(!p_s_where || p_s_where < p_s_text)
		return 0;

	while(p_s_where != p_s_text && (*p_s_where == '\n' || *p_s_where == '\r'))
		-- p_s_where;
	// skip \n\r in case we were given end-of-line characters

	const char *p_s_line_start = p_s_where;
	while(p_s_line_start != p_s_text) {
		if(*(p_s_line_start - 1) == '\n' || *(p_s_line_start - 1) == '\r')
			break;
		-- p_s_line_start;
	}
	// find line start

	int n_column = 0;
	for(const char *p_ch = p_s_line_start; p_ch <= p_s_where; ++ p_ch) {
		if(*p_ch == '\t') {
			n_column += 4;
			n_column -= n_column % 4;
		} else
			++ n_column;
	}
	// count columns, with correct tabs

	return n_column + 1;
}

#define _LineCol_ n_Line(r_p_s_src, p_s_str), n_Column(r_p_s_src, p_s_str)

inline void _SkipWhite_(const char *&x)
{
	while(isspace(*(x)))
		++ x;
}

inline void _SkipIdent_(const char *&x)
{
	if(*x == '_' || (*x >= 'a' && *x <= 'z') || (*x >= 'A' && *x <= 'Z'))
		++ x;
	while(*x == '_' || (*x >= 'a' && *x <= 'z') ||
	   (*x >= 'A' && *x <= 'Z') || (*x >= '0' && *x <= '9'))
		++ x;
	// [_a-zA-Z][_a-zA-Z0-9]*
}

bool PatchTemplateTypenames(char *&r_p_s_src, bool &r_b_touched)
{
	char *p_s_str = r_p_s_src;
	while(p_s_str && (p_s_str = (char*)strstr(p_s_str, "template"))) {
		p_s_str += strlen("template");
		_SkipWhite_(p_s_str);
		if(*p_s_str != '<')
			continue;
		// may not be template syntax at all! (template_variable = 10)

		++ p_s_str; // skip '<'
		_SkipWhite_(p_s_str);
		if(*p_s_str == '>')
			continue; // template <> is legal

		while(true) {
			_SkipWhite_(p_s_str);
			if(!strncmp(p_s_str, "typename", strlen("typename"))) {
				printf("\npatch: template <typename _Ty> fix (ln %d. col %d)", _LineCol_);
				r_b_touched = true;

				memcpy(p_s_str, "class   ", strlen("typename"));
				// could omit spaces, but would have to copy whole string and thet would be slow

				p_s_str += strlen("typename"); // skip class
			} else
				_SkipIdent_(p_s_str); // skip class
			_SkipWhite_(p_s_str); // skip space
			_SkipIdent_(p_s_str); // skip classname

			_SkipWhite_(p_s_str);
			if(*p_s_str == '>')
				break;
			else if(*p_s_str == ',') {
				++ p_s_str; // skip ','
				continue;
			} else if(*p_s_str == '=') {
				++ p_s_str; // skip '='

				while(*p_s_str && *p_s_str != ',' && *p_s_str != '>') {
					if(*p_s_str == '<') {
						fprintf(stderr, "\nerror: complicated template "
							"syntax (ln %d. col %d)\n", _LineCol_);
						return false;
					}
					++ p_s_str;
				}
				if(*p_s_str && *p_s_str == ',') {
					++ p_s_str;
					continue;
				} else if(*p_s_str && *p_s_str == '>') {
					++ p_s_str;
					break;
				} else {
					fprintf(stderr, "\nerror: sick template "
						"syntax (ln %d. col %d)\n", _LineCol_);
					return false;
				}
				// skip '= vlaue'
			} else {
				fprintf(stderr, "\nwarning: do not understand template arg list "
					"syntax (ln %d. col %d)\n", _LineCol_);
				break;
			}
		}
	}
	// replace any  template <typename _Ty0, typename _Ty1, .. typename _Tyn>
	// with			template <class    _Ty0,    class _Ty1, ..    class _Tyn>
	// (doesn't reallocate the string, just keep it simple and with minimal
	// count of memory writes, hence the whitespace)

	return true;
}

bool PatchMSVCUnknownTypes(char *&r_p_s_src, bool &r_b_touched)
{
	/*const struct {
		const char *p_s_typename, *p_s_typedef;
	} p_type_list[] = {
		"ptrdiff_t", "#ifndef __PTRDIFF_T_DEFINED_\r\n"
					 "#define __PTRDIFF_T_DEFINED_\r\n"
					 "typedef long ptrdiff_t;\r\n"
					 "#endif\r\n"
	};

	for(int i = 0; i < sizeof(p_type_list) / sizeof(p_type_list[0]); ++ i) {
		if(strstr(r_p_s_src, p_type_list[i].p_s_typename) &&
		   !strstr(r_p_s_src, p_type_list[i].p_s_typedef)) {
			printf("\npatch: defining %s (ln 1, col 1)", p_type_list[i].p_s_typename);
			r_b_touched = true;

			char *p_s_new_str;
			if(!(p_s_new_str = new char[strlen(r_p_s_src) +
			   strlen(p_type_list[i].p_s_typedef) + 1])) {
				fprintf(stderr, "error: not enough memory\n");
				return false;
			}
			strcpy(p_s_new_str, p_type_list[i].p_s_typedef);
			strcat(p_s_new_str, r_p_s_src);
			delete[] r_p_s_src;
			r_p_s_src = p_s_new_str;
		}
	}*/

	return true;
}

bool PatchRemoveStructures(char *&r_p_s_src, bool &r_b_touched)
{
	const char *p_struct_list[] = {"rel_ops::"};

	for(int i = 0; i < sizeof(p_struct_list) / sizeof(p_struct_list[0]); ++ i) {
		const char *p_s_str;
		while(p_s_str = strstr(r_p_s_src, p_struct_list[i])) {
			printf("\npatch: removing \'%s\' (ln %d, col %d)", p_struct_list[i], _LineCol_);
			r_b_touched = true;

			char *where = (char*)strstr(r_p_s_src, p_struct_list[i]);
			char *what = where + strlen(p_struct_list[i]);

			memcpy(where, what, strlen(what) + 1);
			// remove it
		}
	}

	return true;
}

bool PatchForEachCycles(char *&r_p_s_src, bool &r_b_touched)
{
	char *p_s_str = r_p_s_src;
	while(p_s_str && (p_s_str = (char*)strstr(p_s_str, "for_each"))) {
		char *p_s_for_each_start = p_s_str;

		p_s_str += strlen("for_each");
		_SkipWhite_(p_s_str);

		if(*p_s_str != '(') {
			fprintf(stderr, "\nwarning: do not understand for_each "
				"syntax (ln %d. col %d)\n", _LineCol_);
			continue;
		}

		_SkipWhite_(p_s_str);
		char *p_s_type = p_s_str, *p_s_type_end = 0;
		_SkipIdent_(p_s_str);
		p_s_type_end = p_s_str;
		if(!isspace(*p_s_str)) {
			fprintf(stderr, "\nwarning: do not understand for_each "
				"syntax (ln %d. col %d)\n", _LineCol_);
			continue;
		}
		// get _type_

		_SkipWhite_(p_s_str);
		char *p_s_name = p_s_str, *p_s_name_end = 0;
		_SkipIdent_(p_s_str);
		p_s_name_end = p_s_str;
		if(!isspace(*p_s_str) && *p_s_str != ',') {
			fprintf(stderr, "\nwarning: do not understand for_each "
				"syntax (ln %d. col %d)\n", _LineCol_);
			continue;
		}
		// get _name_

		_SkipWhite_(p_s_str);
		if(*p_s_str != ',') {
			fprintf(stderr, "\nwarning: do not understand for_each "
				"syntax (ln %d. col %d)\n", _LineCol_);
			continue;
		}
		++ p_s_str;
		// skip ','

		_SkipWhite_(p_s_str);
		char *p_s_list = p_s_str, *p_s_list_end = 0;
		_SkipIdent_(p_s_str);
		p_s_list_end = p_s_str;
		if(!isspace(*p_s_str) && *p_s_str != ')') {
			fprintf(stderr, "\nwarning: do not understand for_each "
				"syntax (ln %d. col %d)\n", _LineCol_);
			continue;
		}
		// get _list_

		_SkipWhite_(p_s_str);
		if(*p_s_str != ')') {
			fprintf(stderr, "\nwarning: do not understand for_each "
				"syntax (ln %d. col %d)\n", _LineCol_);
			continue;
		}
		++ p_s_str;
		// skip ')'

		char *p_s_for_each_end = p_s_str;
		// end of for-each

		int n_ln = n_Line(r_p_s_src, p_s_for_each_start),
			n_col = n_Column(r_p_s_src, p_s_for_each_start);

		printf("\npatch: for_each fix (ln %d. col %d)", n_ln, n_col);
		r_b_touched = true;

		*p_s_for_each_start = 0;
		*p_s_type_end = 0;
		*p_s_name_end = 0;
		*p_s_list_end = 0;
		char p_s_counter[256];
		sprintf(p_s_counter, "_patch_%d_%d_counter_", n_ln, n_col);
		char p_s_once[256];
		sprintf(p_s_once, "_patch_%d_%d_once_", n_ln, n_col);
		// create variable names

		char *p_s_new = p_s_Format("%s"
			"for(unsigned int %s = 0, %s = 1; %s < %s.size(); ++ %s, %s = 1)"
				"for(%s %s = %s[%s]; %s; %s = 0)" "%s",

			r_p_s_src,
			// for(unsigned int %s = 0, %s = 1; %s < %s.size(); ++ %s, %s = 1)
			p_s_counter, p_s_once, p_s_counter, p_s_list, p_s_counter, p_s_once,
			// for(%s %s = %s[%s]; %s; %s = 0)
			p_s_type, p_s_name, p_s_list, p_s_counter, p_s_once, p_s_once,
			p_s_for_each_end);
		if(!p_s_new) {
			fprintf(stderr, "error: not enough memory\n");
			return false;
		}
		delete[] r_p_s_src;
		p_s_str = r_p_s_src = p_s_new; // pointers changed
	}

	// replace	for_each(_type_ _name_, _list_)
	// with:
	/*
	for(int _patch_ln_col_ = 0, _patch_ln_col_helper_ = 0; _patch_ln_col_ < _list_.size(); ++ _patch_ln_col_, _patch_ln_col_helper_ = 0)
		for(_type_ _name_ = _list_[_patch_ln_col_helper_]; _patch_ln_col_helper_ < 1; ++ _patch_ln_col_helper_)
	*/
	// that should do it ;)

	return true;
}

bool PatchFile(const char *p_s_filename)
{
	char *p_s_file;
	if(!(p_s_file = p_s_ReadFile(p_s_filename))) {
		fprintf(stderr, "error: error reading file\n");
		return false;
	}

	{
		int n_size = strlen(p_s_file);
		const char *p_unit_name_list[] = {"", "k", "M"};
		float f_size = n_size;
		int n_unit = 0;
		while(n_unit < 3 && f_size > 1000) {
			f_size /= 1024;
			++ n_unit;
		}
		printf("reading \'%s\' (%.2f %sB)        \r", p_s_filename, f_size, p_unit_name_list[n_unit]);
	}
	// print banner

	bool b_touched = false;
	if(!PatchTemplateTypenames(p_s_file, b_touched) ||
	   !PatchMSVCUnknownTypes(p_s_file, b_touched) ||
	   !PatchRemoveStructures(p_s_file, b_touched) ||
	   !PatchForEachCycles(p_s_file, b_touched)) {
		if(p_s_file)
			delete[] p_s_file;
		fprintf(stderr, "error: error patching file\n");
		return false;
	}
	// apply patches

	if(b_touched) {
		printf("\n");
		FILE *p_fw;
		if(!(p_fw = fopen(p_s_filename, "wb"))) {
			fprintf(stderr, "error: error opening file for writing\n");
			delete[] p_s_file;
			return false;
		}
		if(fwrite(p_s_file, strlen(p_s_file) * sizeof(char), 1, p_fw) != 1) {
			fprintf(stderr, "error: error writing file\n");
			fclose(p_fw);
			delete[] p_s_file;
			return false;
		}
		fclose(p_fw);
	}
	// save patched file

	delete[] p_s_file;
	// cleanup

	return true;
}

void _PatchFile(const TFileInfo &r_t_file)
{
	if(r_t_file.b_directory)
		return;
	// not with those

	const char *p_s_path = r_t_file.p_s_Filename();
	if(!p_s_path) {
		fprintf(stderr, "error: not enough memory\n");
		exit(-1);
	}
	const char *p_s_ext = strrchr(p_s_path, '.');
	if(!p_s_ext)
		return;
	++ p_s_ext;
	if(!strcmpi(p_s_ext, "h") || !strcmpi(p_s_ext, "c") ||
	   !strcmpi(p_s_ext, "hpp") || !strcmpi(p_s_ext, "cpp") ||
	   !strcmpi(p_s_ext, "hxx") || !strcmpi(p_s_ext, "cxx") ||
	   !strcmpi(p_s_ext, "inl") || !strcmpi(p_s_ext, "cc") || !strcmpi(p_s_ext, "moc")) {
		if(!PatchFile(p_s_path)) {
			fprintf(stderr, "error: error patching \'%s\'\n", p_s_path);
			exit(-1);
		}
	}
	// patch c/c++ source files + .moc
}

int main(int n_arg_num, const char *p_arg_list[])
{
	if(n_arg_num != 2) {
		fprintf(stderr, "error: not enough arguments. must pass path to qt.\n");
		fprintf(stderr, "error: example synopsis: qtpatch c:\\qt\\4.3.4\n");
		return -1;
	}

	fprintf(stderr, "Your Qt is about to be patched so it could be compiled with MSVC 6.0\n"
					"This may potentionally result in dammage to any of files in specified\n"
					"directory. Make sure you have backup.\n");

	const char *p_s_qt_dir = p_arg_list[1];
	printf("are you sure you want to patch files in \'%s\'? [y/n]\n", p_s_qt_dir);
#ifndef _DEBUG
	fflush(stdin);
	if(tolower(fgetc(stdin)) != 'y') {
		printf("no ... ok, nevermind\n");
		return 0;
	}
#else
	printf("y\n");
#endif

	if(!CDirTraversal::Traverse(p_s_qt_dir, _PatchFile)) {
		fprintf(stderr, "error: error traversing \'%s\' directory structure\n", p_s_qt_dir);
		return -1;
	}

	printf("\n\nand that's it\n");

	return 0;
}
